/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import picture.math.Term;
import java.io.Serializable;
import java.util.*;
import java.text.*;

/**
 * Implementiert den Koerper <b>C</b> der komplexen Zahlen. 
 *
 * @author Stefan Roehrich
 * @author Manuel Kauers
 * @version 1.0
 */
public final class Complex implements Serializable {

    public static final double epsilon = 1E-10;

    /**
     * Beinhaltet den Ralteil dieser Zahl
     */
    protected final double re;

    /**
     * Beinhaltet den Imaginaerteil dieser Zahl
     */
    protected final double im;

    /**
     * Repraesentiert eine Darstellung der imaginaeren Einheit
     */    
    public static final Complex i = new Complex(0.0, 1.0);

    /**
     * Erzeugt eine neue komplexe Zahl, die die Eins repraesentiert.
     */
    public Complex() {re = 1; im = 0;}

    public Complex(double[] value) {this(value[0], value[1]);}

    /**
     * Erzeugt eine neue komplexe Zahl aus Real- und Imaginaerteil.
     * @param re Realteil
     * @param im Imaginaerteil
     */
    public Complex(double re, double im) {
	this.re = re;
	this.im = im;
    }

    private static String[] var = new String[0];
    private static Complex[] ers = new Complex[0];
    
    /**
     * Erzeugt eine neue komplexe Zahl aus einem Sting, im Fehlerfall 0.
     * @param s String
     */
    public Complex(String s) {
	Complex t;
	try {
	    t = new Term(s).einsetzen(var, ers);
	} catch(Exception e) {
	    this.re = this.im = 0.0;
	    return;
	}
	this.re = t.re;
	this.im = t.im;
    }

    /**
     * Liefert die n-te Einheitswurzel von eins, definiert als 
     * <center>
     * exp(2 <i>pi i / n</i>)
     * </center>
     * @param n Wurzelexponent
     * @return die <i>n</i>-te Einheitswurzel von eins.
     */
    public final Complex createUnitRoot(int n) {
    	return (new Complex(0, -2 * Math.PI / n)).exp();
    }

    /**
     * Liefert die Summe von this und dem Argument.
     * @param b zweiter Summand
     * @return this + b
     */
    public final Complex add(Complex b) {
	return new Complex(re + b.re, im + b.im);
    }

    /**
     * Liefert die komplexe Zahl, die entsteht, wenn man den
     * uebergebenen double zum Realteil von this hinzuzaehlt
     * @param b zweiter Summand
     * @return this + b
     */
    public final Complex add(double b) {
	return new Complex(re + b, im);
    }

    /**
     * Liefert die Differenz von this zum Argument.
     * @param b Minuend
     * @return this - b
     */
    public final Complex sub(Complex b) {
	return new Complex(re - b.re, im - b.im);
    }

    /**
     * Liefert die komplexe Zahl, die entsteht, wenn man
     * den uebergebenen double vom Realteil dieser Zahl 
     * abzieht.
     * @param b Minuend
     * @return this - b
     */
    public final Complex sub(double b) {
	return new Complex(re - b, im);
    }

    /**
     * Liefert das Produkt von this und dem Argument.
     * @param b zweiter Faktor
     * @return this * b
     */
    public final Complex mul(Complex b) {
	return new Complex(re * b.re - im * b.im, im * b.re + re * b.im);
    }

    /**
     * Liefert die komplexe Zahl, die entsteht, wenn man 
     * Realteil und Imaginaerteil dieser Zahl mit dem 
     * uebergebenen double multipliziert
     * @param b zweiter Faktor
     * @return this * b
     */
    public final Complex mul(double b) {
	return new Complex(re * b, im * b);
    }

    /**
     * Liefert den Quotienten von this und dem Argument.
     * @param b Divisor
     * @return this / b
     */
    public final Complex div(Complex b) {
	double nenner = b.re * b.re + b.im * b.im;
	return new Complex((re * b.re + im * b.im) / nenner, (im * b.re - re * b.im) / nenner);
    }

    /**
     * Liefert die komplexe Zahl, die entsteht, wenn man 
     * Realteil und Imaginaerteil dieser Zahl durch den 
     * uebergebenen double teilt
     * @param b Divisor
     * @return this / b
     */
    public final Complex div(double b) {
	return new Complex(re / b, im / b);
    }

    /**
     * Liefert die komplex konjugierte Zahl zu this
     * @return this quer
     */
    public final Complex conjugated() {
	return new Complex(re, -im);
    }

    /**
     * Liefert <tt>exp(this)</tt>, definiert als
     * <center>
     *  exp(<i>a</i> + i <i>b</i>) 
     * = <i>e</i><sup><i>a</i></sup> (cos(<i>b</i>) + i sin(<i>b</i>))
     * </center>
     * @return exp(this)
     */
    public final Complex exp() {
	double exp = Math.exp(re);
	return new Complex(exp * Math.cos(im), exp * Math.sin(im));
    }

    /**
     * Liefert this^n, wobei n natuerlich ist
     * @param n Exponent
     * @return this^n
     */
    public final Complex power(int n) {
	if (n < 0)
	    return (new Complex()).div(this).power(-n);
	else if (n == 0)
	    return new Complex(); 
	else if (n == 1)
	    return this;

	Complex a = new Complex();
	int m;

	for(m = 1; m <= n; m <<= 1) ; // leere Schleife

	m >>= 1;

	for(int i = m; i > 0; i >>= 1)
	    a = ((i&n) == 0) ? a.mul(a) : mul(a.mul(a));

	return a;
    }

    /**
     * gibt den Realteil dieser Zahl als double zurueck
     * @return Re(this)
     */
    public final double getRealPart() {return re;}
    
    /**
     * gibt den Imaginaerteil dieser Zahl als double zurueck
     * @return Im(this)
     */
    public final double getImaginaryPart() {return im;}

    /**
     * gibt das Argument dieser Zahl, also den Winkel in der 
     * Polarkoordinatendarstellung, als double zurueck.
     * Im Fall <tt>this = 0</tt> (d.h. <tt>Arg(this)</tt> nicht def.) liefert diese 
     * Methode Null. Mit <tt>abs(this) = 0</tt> f&uuml;hrt dies wegen
     * <center>
     *  <tt>this = abs(this) * exp(i * Arg(this)) = 0 * 0 = 0</tt>
     * </center>
     * zu einem sinnvollen Ergebnis (im Gegensatz etwa zu <tt>Arg(0) := NaN</tt>).
     * @return Arg(this)
     */
    public final double getPolarAngle() {

	double arg = Double.NaN;
	
	if (im == 0 && re == 0)
	    arg = 0;
	else if (im == 0 && re < 0)
	    arg = Math.PI;
	else 
	    arg = ((im < 0) ? -1 : 1) * Math.acos(re / getAbsolutValue());

	return arg;
    }


    /**
     * gibt den Absolutbetrag dieser Zahl als double zurueck
     * @return abs(this)
     */
    public final double getAbsolutValue() {

	return Math.sqrt(re * re + im * im);
    }    

    public final boolean equals(Object o) {

	if (!(o instanceof Complex))
	    return false;

	Complex c = (Complex) o;
	return (Math.abs(re - c.re) < epsilon && Math.abs(im - c.im) < epsilon);
    }

    /**
     * Liefert eine String-Repraesentation dieser Zahl
     * @return String-Repraesentation
     */
    public final String toString2() {

	/*
	  Die Konversion ist wie folgt spezifiziert: Ist abs(this) < epsilon, 
	  wird "0" angegeben. Ist abs(re) oder abs(im) < epsilon, so wird nur
	  das jeweilige andere angegeben. Jede Zahl wird dargestellt in der
	  Form " 1.111e+11" oder "-1.111e-11". Durch fuehrende Nullen oder 
	  Leerzeichen wird gewaehrleistet, dass jede Zahl einen 10 Zeichen langen
	  String generiert. Ist der Exponent 0, so werden anstelle von "e+00" 
	  weitere Nachkommastellen eingefuegt.
	  Imaginaere Zahlen sind zwei Zeichen laenger durch
	  nachgestelltes " i", etwa " 1.0000000 i". 
	  Komplexe Zahlen sind 25 Zeichen lang, etwa "-6.543e+21 +  1.234e-56 i"
	 */
	if (getAbsolutValue() < epsilon)
	    return "0";
	else if (Math.abs(im) < epsilon)
	    return "" + string(re);
	else if (Math.abs(re) < epsilon)
	    return "" + string(im) + " i";
	else
	    return string(re) + " + " + string(im) + " i";
    }

    private static NumberFormat nf = NumberFormat.getInstance(Locale.US);

    static {
	nf.setGroupingUsed(false);
	nf.setMaximumFractionDigits(8);
	// nf.setMaximumIntegerDigits(5);
    }

    /**
     * Liefert eine String-Repraesentation dieser Zahl
     * @return String-Repraesentation
     */
    public final String toString() {
	String s = "";

	if(Math.abs(re) <= epsilon && Math.abs(im) <= epsilon)
	    return "0";
	if(Math.abs(re) > epsilon)
	    s = nf.format(re) + (im>epsilon ? "+" : "");
	if(Math.abs(im) > epsilon)
	    s = s + nf.format(im) + "*i";

	return s;
    }

    private String string(double d) {

	if (d == 0)
	    return "0";

	String eingabe = "" + d;
	eingabe = eingabe.toLowerCase();

	// Wir parsen eingabe und zerlegen es in 
	String vzm = " "; // vorzeichen der Mantisse
	String vkm = ""; // vorkommateil der Mantisse
	String nkm = ""; // nachkommateil der Mantisse
	String vze = " "; // vorzeichen des Exponenten
	String exp = ""; // Betrag des Exponenten.	
	
	final int VZ = 0; 	final int VKM = 1;
	final int NKM = 2;	final int VZE = 3;
	final int EXP = 4;

	int status = VZ;

	// entferne Leerzeichen
	String e = "";
	for (int i = 0; i < eingabe.length(); i++) {

	    char c = eingabe.charAt(i);
	    if (c != ' ')
		e += c;
	}

	eingabe = e;	
	for (int i = 0; i < eingabe.length(); i++) {

	    char c = eingabe.charAt(i);

	    switch (status) {
	    case VZ: 
		if (c == '-') {vzm = "-"; status++; break;} 
		if (c == '+') {status++; break;}
		status++; // kein break;
	    case VKM:
		if (c == 'e') {status += 2; break;}
		if (c == '.' || c == ',') {status++; break;}
		vkm += c; break;		
	    case NKM:
		if (c == 'e') {status++; break;}
		nkm += c; break;
	    case VZE:
		if (c == '-') {vze = "-"; status++; break;}
		if (c == '+') {status++; break;}
		status++; // kein break;
	    case EXP:
		exp += c; break;
	    }
	}

	// fuelle Nachkommastellen auf, bis es mindestens zehn sind
	nkm += "0000000000";

	// falls die Mantisse mehr als eine Vorkommastelle hat, aendere 
	// dies zugunsten des Exponenten

	if (exp.length() == 0) exp = "0";
	int expI = Integer.parseInt(exp);
	if (vze.equals("-")) expI *= -1;

	if (vkm.length() == 0) {
	    // Bsp: .5e1 -> 5.0e0
	    vkm = "" + nkm.charAt(0);
	    nkm = nkm.substring(1);
	    expI--;
	}

	if (vkm.length() > 1) {
	    // Bsp: 12e0 -> 1.2e1
	    nkm = vkm.substring(1) + nkm;
	    expI += vkm.substring(1).length();
	    vkm = "" + vkm.charAt(0);
	}
	
	exp = "00" + Math.abs(expI);
	vze = " "; if (expI < 0) vze = "-";
	
	// setze die Ausgabe zusammen
	String ausgabe = vzm + vkm + ".";
	if (expI == 0) {
	    ausgabe += nkm.substring(0, 8);
	} else {
	    ausgabe += nkm.substring(0, 4) + "e" + vze;
	    ausgabe += exp.substring(exp.length() - 2, exp.length());
	}

	return ausgabe;
    }

    public static void main(String[] args) {
	System.out.println(new Complex(""));
	System.out.println(new Complex("1"));
	System.out.println(new Complex("i"));
	System.out.println(new Complex("1+5*i"));
	System.out.println(new Complex("i+5"));
	System.out.println(new Complex("sqrt(2)+3*i"));
	System.out.println(new Complex("3.4123"));
    }
}

