/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import java.awt.*;
import java.awt.image.*;
import javax.swing.JFrame;

import picture.gui.*;
import picture.filter.*;
import picture.filetype.*;

/**
 * Diese Klasse repraesentiert ein Bild in Form eines Bitmaps mit reellwertigen
 * Pixeldaten. Als Huellklasse f&uuml;r einen zweidimensionalen <tt>double</tt>-Array
 * findet diese Klasse Anwendung als Bildebene in Picture.
 * 
 * @version 2.0 (c) 2000
 * @author Stefan Roehrich
 * @author Manuel Kauers
 */
public final class Layer implements PictureConstants {

    /**
     * Enthaelt die Pixeldaten dieses Pictures. Dabei bezeichnet die erste Komponente
     * die Zeile, die zweite die Spalte. 
     */
    public double[][] data;

    /** Bezeichnet die Breite dieses Pictures in Pixeln */
    protected int width;

    /** Bezeichnet die Hoehe dieses Pictures in Pixeln */
    protected int height;

    private transient Layer preview = null;
    private transient Image img = null;

    private double min = Double.POSITIVE_INFINITY;
    private double max = Double.NEGATIVE_INFINITY;
    private boolean fixedBounds = false;

    private static final double SQRT2 = Math.sqrt(2);

    /**
     * Zum internen Gebrauch. Erzeugt ein Bild der Groesse 2 x 2 
     */
    public Layer() {this(new double[][]{{0, 1}, {2, 3}});}

    /**
     * Erzeugt ein neues Bild.
     * 
     * @param data Bilddaten.
     */
    public Layer(double[][] data) {

	this.data = data;
	width = data.length;
	height = data[0].length;
    }

    /**
     * Erzeugt ein neues Bild. Die Schranken werden dabei explizit mit angegeben.
     * Es ist darauf zu achten, dass min und max so gewaehlt sind, dass alle 
     * Bilddaten dazwischen liegen. <p>
     * Sinnvoll ist die Angabe eigener Grenzen z.B. zur Erzeugung von konstanten,
     * nichtschwarzen Bilder oder allgemein zur Vermeidung der farblichen
     * Autoskalierung. 
     *
     * @param data Bilddaten.
     * @param min Postulierte untere Schranke der Bilddaten
     * @param max Postulierte obere Schranke der Bilddaten
     */
    public Layer(double[][] data, double min, double max) {

	this(data);

	if (min > max) {
	    double c = min; min = max; max = c;
	}

	this.min = min; this.max = max;
	fixedBounds = true;
    }

    /**
     * Erzeugt ein neues quadratisches Bild der Kantenlaenge <tt>b</tt>. 
     * Das Bild ist konstant 0.
     * 
     * @param b
     */
    public Layer(int b) {

	this.data = new double[b][b];
	width = b; height = b;
    }

    /**
     * Erzeugt ein Bild, das durch zyklische Verschiebung dieses Bildes in 
     * Hoehe und Breite entsteht.
     * @param width Verschiebung in <tt>x</tt>-Richtung
     * @param height Verschiebung in <tt>y</tt>-Richtung
     */
    public Layer getShiftedInstance(int width, int height) {

	Layer p = createNewInstance();

	int w0 = width; int h0 = height;
	int w = this.width; int h = this.height;

	for (int i = 0; i < this.width; i++)
	    for (int j = 0; j < this.height; j++)
		p.data[i][j] = data[(i + w0) % w][(j + h0) % h];

	return p;
    }

    /**
     * gibt das kleinste quadratische Bild zurueck mit Kantenlaenge 2^<i>n</i>,
     * in das dieses Bild hineinpasst. Dieses Bild wird in dem neu konstruierten
     * Bild zentriert dargestellt. Die Raender werden mit 0 aufgefuellt.
     * @return naechstgroesseres quadratisches Zweierpotenzbild
     */
    public Layer getEmbeddedInstance() {

	int w = width;
	if (height > w) w = height;
	
	int h = 1;
	while (w > h) h *= 2;

	Layer p = createNewInstance("", h, h);
	for (int i = 0; i < h; i++)
	    for (int j = 0; j < h; j++)
		p.data[i][j] = 0;

	int wOffset = (h - width) / 2;
	int hOffset = (h - height) / 2;

	for (int i = 0; i < width; i++)
	    for (int j = 0; j < height; j++)
		p.data[wOffset + i][hOffset + j] = data[i][j];

	p.width = h; p.height = h;
	return p;
    }

    /**
     * Erzeugt eine in der Groesse skalierte Version dieses Bildes. 
     * @param width Breite des neuen Bildes
     * @param height Hoehe des neuen Bildes
     * @return eine Projektion dieses Bildes auf die Groesse <tt>width</tt> mal <tt>height</tt>.
     */
    public Layer getScaledInstance(int width, int height) {

	Layer p = createNewInstance("Preview", width, height);

	for (int i = 0; i < p.width; i++)
	    for (int j = 0; j < p.height; j++)
		p.data[i][j] = data[i * this.width / p.width][j * this.height / p.height];

	return p;
    }

    /**
     * Erzeugt ein Layer wie dieses, in dem jedes Pixel mit dem angegebenen 
     * double-Wert multipliziert wurden.
     *
     * @param w Gewicht
     * @return mit <tt>w</tt> gewichtete Version dieses Layers
     */
    public Layer getWeightedInstance(double w) {

	Layer p = createNewInstance("", width, height);

	for (int x = 0; x < width; x++)
	    for (int y = 0; y < height; y++)
		p.data[x][y] = w * data[x][y];

	if (fixedBounds) {
	    p.fixedBounds = true;
	    p.min = min;
	    p.max = max;
	}
	return p;
    }

    /**
     * Erzeugt ein Vorschaubild aus diesem Objekt. Fuer Dialogfenster kann diese 
     * Methode zur Erzeugung einer Kopie dieses Pictures in der Standardgroesse 
     * 64 mal 64 verwendet werden. 
     * @return Ein neues Layer, das dieses Bild in der Normgroesse 64 mal 64 enthaelt.
     */
    public Layer getPreview() {

	if (preview != null) return preview;

	try {
	    if (width == height && DefaultTransformation.isPowerOfTwo(width) && height > 64) {

		Wavelet w = Wavelet.getHaarWavelet(Wavelet.ANALYSIS_WITHOUT_DETAILS);
		double [][] in = data; 

		Layer p = this;
		while (p.width > 64) p = w.of(null, p);
		preview = p;

	    } else {
		preview = getScaledInstance(64, 64);
	    }

	} catch (Exception e) {}

	return preview;
    }

    /**
     * Gibt die Breite dieses Bildes in Pixeln zurueck 
     * @return Breite dieses Bildes.
     */
    public int getWidth() {return width;}

    /**
     * Gibt die Hoehe dieses Bildes in Pixeln zurueck
     * @return Hoehe dieses Bildes
     */
    public int getHeight() {return height;}

    /**
     * Gibt den double-Wert des Pixels (x, y) in der aktuellen Praesentation an.
     * @param x zu untersuchende <tt>x</tt>-Koordinate
     * @param y zu untersuchende <tt>y</tt>-Koordinate
     * @return Grauwert in <tt>(x, y)</tt> in der aktuellen Praesentation.
     */
    public double getValue(int x, int y) {

	if (min == Double.POSITIVE_INFINITY || max == Double.NEGATIVE_INFINITY)
	    berechneMinimumUndMaximum();

	if (x < 0 || y < 0 || x >= width || y >= height) return 0;

	return Math.max(min, Math.min(max, data[x][y]));
    }

    /**
     * Gibt den maximalen logischen Grauwert in diesem Bild wider.
     *
     * @return maximaler Grauwert.
     */
    public double getMaximum() {

	if (max == Double.NEGATIVE_INFINITY) berechneMinimumUndMaximum();
	return max;
    }

    /**
     * Setzt den maximalen logischen Grauwert.
     *
     * @param max maximaler Grauwert
     */
    public void setMaximum(double max) {
	this.max = max;
    }

    /**
     * Gibt den minimalen logischen Grauwert in diesem Bild wider.
     *
     * @return minimaler Grauwert.
     */
    public double getMinimum() {

	if (min == Double.POSITIVE_INFINITY) berechneMinimumUndMaximum();
	return min;
    }

    /**
     * Setzt den minimalen logischen Grauwert.
     *
     * @param min minimaler Grauwert
     */
    public void setMinimum(double min) {
	this.min = min;
    }

    /**
     * Berechnet Minimum und Maximum der Bilddaten.
     *
     * @return maximaler Grauwert in der aktuellen Ansicht.
     */
    private synchronized void berechneMinimumUndMaximum() {

	max = Double.NEGATIVE_INFINITY; min = Double.POSITIVE_INFINITY;

	for (int x = 0; x < width; x++)
	    for (int y = 0; y < height; y++) {
		if (max < data[x][y]) max = data[x][y];
		if (min > data[x][y]) min = data[x][y];
	    }
    }

    /**
     * Erzeugt einen Informationsstring zum Pixel an der Stelle (<tt>x</tt>, <tt>y</tt>).
     * @param x <tt>x</tt>-Koordinate
     * @param y <tt>y</tt>-Koordinate
     * @return Info-String zum Pixel (<tt>x</tt>, <tt>y</tt>).
     */
    public String getInfo(int x, int y) {

	if (x < 0 || y < 0 || x >= width || y >= height)
	    return "";

	return "" + data[x][y];
    }

    /**
     * Liefert die Pixelverteilung dieses Pictures
     * zurck. Dabei wird das Intervall [min, max] in eine
     * quidistante Zerlegung der Lnge <tt>l</tt> zerlegt und ein
     * <tt>int[]</tt>-Feld der entsprechenden Lnge mit den jeweiligen
     * Hufigkeiten zurckgegeben.
     *
     * @param l Zerlegungslnge
     */
    public int[] histogram(int l) {

	int[] h = new int[l];
	int i;
	
	for(int x = 0; x < width; x++)
	    for(int y = 0; y < height; y++) {
		i = (int) ((data[x][y] - min) / (max - min) * (l - 1));
		if (i < 0) i = 0;
		if (i > l - 1) i = l - 1;
		h[i]++;
	    }
	
	return h;
    }

    private Layer createNewInstance() {

	return createNewInstance("", width, height);
    }

    private Layer createNewInstance(String name, int width, int height) {

	return createNewInstance(name, new double[width][height]);
    }

    /**
     * Erzeugt eine Kopie dieses Bildes als Instanz der Unterklasse, mit der
     * dieses Bild erzeugt wurde. Alle Daten werden uebernommen, ausser <tt>width</tt>,
     * <tt>height</tt>, <tt>data</tt> und <tt>name</tt>, diese werden aus den
     * Parametern uebernommen.
     * <p>
     * @param name: Name des neuen Bildes (Nun Unfug, da Layer keine Namen mehr haben...)
     * @param data: Inhalt des neuen Bildes.
     * @return Layer wie <tt>this</tt>, nur mit <tt>data</tt> als Inhalt.
     */
    public Layer createNewInstance(String name, double[][] data) {

	try {
	    Layer p = (Layer) (this.getClass().newInstance());
	    p.width = data.length;
	    p.height = data[0].length;
	    p.data = data;
	    return p;
	} catch (Exception ign) {}       
	return null;
    }
}
