/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

import picture.*;
import picture.filter.*;

public class ColorKLTPane extends DialogPane {

    private static int threadNumber = 0;

    private class EigenValueArray extends JScrollPane 
	implements Runnable, ChangeListener, ActionListener {

	boolean kltOrNot;
	private EigenValue[] array;
	JPanel comps = new JPanel(new GridLayout(0, 1));
	Thread update = new Thread(this);
	long lastCall = System.currentTimeMillis();
	boolean killThread = false;
	double[] v;
	boolean blocked = true;

	public EigenValueArray(int number, boolean klt) {

	    this(number, new JComponent[number], klt);
	}

	public EigenValueArray(int number, JComponent[] charts, boolean klt) {

	    super(VERTICAL_SCROLLBAR_AS_NEEDED, HORIZONTAL_SCROLLBAR_NEVER);
	    update.start();
	    setViewportView(comps);
	    setBorder(new BevelBorder(BevelBorder.LOWERED));
	    this.kltOrNot = klt;
	    array = new EigenValue[number];

	    array[number - 1] = (charts[number - 1] == null) ?
		new EigenValue(1, 2) : new EigenValue(1, 2, charts[number - 1]);

	    for (int i = number - 2; i >= 0; i--) 
		array[i] = (charts[i] == null) ? 
		    new EigenValue(1, array[i + 1]) : new EigenValue(1, array[i + 1], charts[i]);

	    for (int i = 0; i < array.length; i++) {
		comps.add(array[i]);
		array[i].addChangeListener(this);
	    }
	}

	public Dimension getPreferredSize() {

	    array[0].validate();
	    Dimension d = array[0].getPreferredSize();
	    d = new Dimension(d.width + getVerticalScrollBar().getPreferredSize().width,
			      3 * d.height + 5);
	    if (d.width < 100) d.width = 250;
	    if (d.height < 100) d.height = 100;
	    return d;
	}

	public EigenValue getEigenValue(int i) {return array[i];}

	public synchronized void kill() {killThread = true; blocked = false; notify();}

	public void run() {

	    try {
		synchronized(this) {wait();}
	    } catch(InterruptedException e) {
	    }

	    while (!killThread) 
		try {
		    abarbeiten();
		} catch (InterruptedException ex) {
		} catch(PictureException ex) {
		    GUIHelp.error(ex);
		    blocked = true; v = null;
		}
	}

	private void abarbeiten() throws PictureException, InterruptedException {

	    Picture p;
	    
	    while (blocked || v == null || System.currentTimeMillis() - lastCall < 200)
		if (killThread) return;
		else Thread.yield();

	    if (kltOrNot) {
		p = klt.kltToStandard(kltPrevSource, v);
		kltPreview.setPictureModel(p);
		kltPreview.repaint();
	    } else {
		p = sourcePrev.getWeightedInstance(v);
		rgbPreview.setPictureModel(p);
		rgbPreview.repaint();
	    }
	    
	    blocked = true; v = null; Thread.sleep(50);
	    synchronized(this) {
		while(blocked)
		    wait();
	    }
	}

	public synchronized void stateChanged(ChangeEvent e) {

	    v = new double[array.length];
	    for (int i = 0; i < v.length; i++)
		v[i] = array[i].getValue();
	    lastCall = System.currentTimeMillis();
	    blocked = false;
	    notify();
	}

	public void actionPerformed(ActionEvent e) {

	    String k = e.getActionCommand();
	    if (k.equals("OK")) {
		closingCode = OK;
		dispose();
	    } else if (k.equals("Abbrechen")) {
		closingCode = ABBORD;
		dispose();
	    }
	}
    }

    private class Chart extends JComponent {

	private final Color c;

	public Chart(Color c) {this.c = c; setBorder(new BevelBorder(BevelBorder.LOWERED));}

	public Dimension getPreferredSize() {return new Dimension(12, 12);}

	public void paintComponent(Graphics g) {

	    g.setColor(c);
	    g.fillRect(0, 0, getWidth(), getHeight());
	}
    }

    private Picture source;
    private Picture sourcePrev;
    private Picture kltSource;
    private Picture kltPrevSource;
    private LayerKLT klt;
    private JPicture kltPreview;
    private JPicture rgbPreview;
    private JComboBox comboBox = new JComboBox();

    private EigenValueArray kltEigenVals;
    private EigenValueArray rgbEigenVals;

    public ColorKLTPane(Frame owner, Picture p) throws PictureException {

	super(owner, "Karhunen-Loeve-Transformation auf dem Farbraum");

	if (p.getNumberOfLayers() != 3)
	    throw new PictureException("Das Bild ist kein Farbbild");

	source = p;
	sourcePrev = p.getPreview();
	klt = new LayerKLT(source);
	kltSource = klt.standardToKlt(source);
	kltPrevSource = kltSource.getPreview();

	addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		rgbEigenVals.kill(); kltEigenVals.kill();
	    }
	});

	JComponent[] charts = new Chart[3];
	charts[0] = new Chart(Color.red);	charts[1] = new Chart(Color.green); 
	charts[2] = new Chart(Color.blue);
	rgbEigenVals = new EigenValueArray(3, charts, false);

	double[][] evs = klt.getEigenVectors();
	for (int i = 0; i < 3; i++) 
	    charts[i] = new Chart(vectorToColor(evs[i]));
	kltEigenVals = new EigenValueArray(3, charts, true);

	Picture prev = p.getPreview();
	kltPreview = new JPicture(prev); rgbPreview = new JPicture(prev);

	Container c = getContentPane();	c.setLayout(new BorderLayout());
	
	JPanel kltPane = new JPanel(new BorderLayout(10, 10));
	JPanel rgbPane = new JPanel(new BorderLayout(10, 10));
	JPanel origPane = new JPanel(new BorderLayout(10, 10));
	JPanel buttons = new JPanel(new GridLayout(1, 5, 10, 10));

	JButton ok = new JButton("OK");
	JButton abbrechen = new JButton("Abbrechen");

	comboBox.addItem("Std-Basis");
	comboBox.addItem("KLT-Basis");
	comboBox.setSelectedItem("KLT-Basis");

	c.add("South", buttons);
	c.add("East", kltPane);
	c.add("West", rgbPane);
	c.add("North", origPane);

	buttons.setBorder(new EmptyBorder(10, 10, 10, 10));
	buttons.add(ok);
	buttons.add(abbrechen);
	buttons.add(new JPanel());
	buttons.add(new JPanel());

	JPanel optionPane = new JPanel(new GridLayout(0, 1));
	optionPane.add(new JLabel("Zur Hervorhebung der unterschiedlichen Eigenschaften sind"));
	optionPane.add(new JLabel("Manipulationen an Standard- und KLT-Basis " + 
				  "nebeneinander mglich."));
	JPanel letzteZeile = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
	letzteZeile.add(new JLabel("Die Schaltflche OK bezieht sich auf die Darstellung zur "));
	letzteZeile.add(comboBox);

	optionPane.add(letzteZeile);

	origPane.setBorder(new EmptyBorder(10, 10, 10, 10));
	origPane.add("West", new JPicture(prev));
	origPane.add("Center", optionPane);

	JPanel rgbPaneTop = new JPanel(new BorderLayout());
	rgbPaneTop.add("East", rgbPreview);

	rgbPane.setBorder(new TitledBorder(new EtchedBorder(), "Standardbasis"));
	rgbPane.add("North", rgbPaneTop);
	rgbPane.add("Center", rgbEigenVals);
	
	JPanel kltPaneTop = new JPanel(new BorderLayout());
	kltPaneTop.add("East", kltPreview);
	//kltPaneTop.add("West", new JPicture(kltSource.getPreview()));

	kltPane.setBorder(new TitledBorder(new EtchedBorder(), "optimale KLT-Basis"));
	kltPane.add("North", kltPaneTop);
	kltPane.add("Center", kltEigenVals);

	ok.addActionListener(kltEigenVals);
	abbrechen.addActionListener(kltEigenVals);
	
	prepareWindow();
    }

    public Picture getResultingPicture() throws PictureException {

	if (comboBox.getSelectedItem().equals("KLT-Basis")) 
	    return getResultKLT();
	
	return getResultStd();
    }

    public Picture getResultKLT() throws PictureException {

	double[] v = new double[3];

	for (int i = 0; i < 3; i++)
	    v[i] = kltEigenVals.getEigenValue(i).getValue();

	return klt.kltToStandard(kltSource, v);
    }

    public Picture getResultStd() throws PictureException {

	double[] v = new double[3];

	for (int i = 0; i < 3; i++)
	    v[i] = rgbEigenVals.getEigenValue(i).getValue();

	return source.getWeightedInstance(v);
    }

    private static final Color vectorToColor(double[] d) {

	if (d.length != 3) return Color.black;
	double k = Math.min(d[0], Math.min(d[1], d[2]));
	if (k < 0)
	    for (int i = 0; i < 3; i++)
		d[i] -= k;
	k = d[0] * d[0] + d[1] * d[1] + d[2] * d[2];
	k = Math.sqrt(k);
	if (k == 0) return Color.black;

	return new Color((float) (d[0] / k), (float) (d[1] / k), (float) (d[2] / k));
    }
}
