/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Input stream which transforms its data via arithmetic encoding.
 *
 * This implementation is based on ideas of <a
 * href="http://www.dogma.net/markn/articles/bwt/bwt.htm">c source
 * code from Mark Nelson</a>.
 *
 * This stream performs an order-0 adaptive arithmetic encoding
 * function. Based on C code from 1987 CACM article by Witten, Neal,
 * and Cleary.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class ArithInputStream extends FilterInputStream implements ArithConstants {

/* BIT INPUT ROUTINES. */

/* THE BIT BUFFER. */

    private int buffer;                     /* Bits waiting to be input                 */
    private int bits_to_go;                 /* Number of bits still in buffer           */
    private int garbage_bits;               /* Number of bits past end-of-file          */

    private long value;        /* Currently-seen code value                */
    private long low, high;    /* Ends of current code region              */

/* TRANSLATION TABLES BETWEEN CHARACTERS AND SYMBOL INDEXES. */

    private int[] char_to_index = new int[NO_OF_CHARS];	/* To index from character          */
    private int[] index_to_char = new int[NO_OF_SYMBOLS+1];	/* To character from index    */

/* ADAPTIVE SOURCE MODEL */

    private int[] freq = new int[NO_OF_SYMBOLS+1];	/* Symbol frequencies                       */
    private int[] cum_freq = new int[NO_OF_SYMBOLS+1];	/* Cumulative symbol frequencies            */

    private boolean first = true;

/* INITIALIZE BIT INPUT. */

    private void start_inputing_bits() {
	bits_to_go = 0;                             /* Buffer starts out with   */
	garbage_bits = 0;                           /* no bits in it.           */
    }


/* INPUT A BIT. */

    private final int input_bit() throws IOException {
	int t;
	if (bits_to_go==0) {                        /* Read the next byte if no */
	    buffer = in.read();                   /* bits are left in buffer. */
	    if (buffer == -1) {
		garbage_bits++;                      /* Return arbitrary bits*/
		if (garbage_bits>CODE_VALUE_BITS-2) {   /* after eof, but check */
		    throw new IOException("Bad input file"); /* for too many such.   */
		}
	    }
	    bits_to_go = 8;
	}
	t = buffer&1;                               /* Return the next bit from */
	buffer >>= 1;                               /* the bottom of the byte.  */
	bits_to_go--;
	return t;
    }


    /**
     * Initialize and set input stream. */
    public ArithInputStream(InputStream in) {
	super(in);
    }

    /**
     * Reads and transforms a byte from the underlying input
     * stream.
     *
     * @return tranformed byte or -1 for end of file
     * @exception IOException
     */
    public int read() throws IOException {
	if(first) {
	    start_model();			/* Set up other modules.    */
	    start_inputing_bits();
	    start_decoding();
	    first = false;
	}

        int ch;
	int symbol;
        symbol = decode_symbol(cum_freq);       /* Decode next symbol.      */
	if (symbol==EOF_SYMBOL)
	    return -1;          /* Exit loop if EOF symbol. */
        ch = index_to_char[symbol];             /* Translate to a character.*/
        update_model(symbol);                   /* Update the model.        */
	return ch;
    }

    /**
     * Reads and transforms a byte array from the underlying input
     * stream. The data is read byte for byte via read() from the
     * underlying input stream.
     *
     * @param b where to write the bytes to
     * @param off starting offset
     * @param len how many bytes to read
     * @return number of bytes bytes actually read
     * @exception IOException */
    public int read(byte[] b, int off, int len) throws IOException {
	int t, i;
        for(i = off; i < off+len; i++)
            if((t = read()) == -1)
		break;
	    else
		b[i] = (byte)t;
	return i;
    }

    /**
     * Do nothing.
     */
    public void mark(int readlimit) {
    }

    /**
     * Do nothing.
     */
    public void reset() throws IOException {
    }

    /**
     * @return false
     */
    public boolean markSupported() {
	return false;
    }

    /**
     * Skips bytes.
     *
     * @param n how many bytes to skip.
     * @return number of bytes actually skipped
     * @exception IOException */
    public long skip(long n) throws IOException {
	long i;
	int t;
	for(i = 0; i < n; i++)
	    if((t = read()) == -1)
		break;
	return i;
    }

/* ARITHMETIC DECODING ALGORITHM. */

/* CURRENT STATE OF THE DECODING. */

/* START DECODING A STREAM OF SYMBOLS. */

    private void start_decoding() throws IOException {
	int i;
	value = 0;                                  /* Input bits to fill the   */
	for (i = 1; i<=CODE_VALUE_BITS; i++) {      /* code value.              */
	    value = 2*value+input_bit();
	}
	low = 0;                                    /* Full code range.         */
	high = TOP_VALUE;
    }


/* DECODE THE NEXT SYMBOL. */

    private int decode_symbol(int cum_freq[]) throws IOException {
	long range;                 /* Size of current code region              */
	int cum;                    /* Cumulative frequency calculated          */
	int symbol;                 /* Symbol decoded                           */
	range = (long)(high-low)+1;
	cum = (int)                                 /* Find cum freq for value. */
	    ((((long)(value-low)+1)*cum_freq[0]-1)/range);
	for (symbol = 1; cum_freq[symbol]>cum; symbol++) ; /* Then find symbol. */

	high = low +                                /* Narrow the code region   */
	    (range*cum_freq[symbol-1])/cum_freq[0]-1; /* to that allotted to this */
	low = low +                                 /* symbol.                  */
	    (range*cum_freq[symbol])/cum_freq[0];
	
	for (;;) {                                  /* Loop to get rid of bits. */
	    if (high<HALF) {
		/* nothing */                       /* Expand low half.         */
	    }
	    else
		if (low>=HALF) {                   /* Expand high half.        */
		    value -= HALF;
		    low -= HALF;                        /* Subtract offset to top.  */
		    high -= HALF;
		}
		else
		    if (low>=FIRST_QTR                 /* Expand middle half.      */
			&& high<THIRD_QTR) {
			value -= FIRST_QTR;
			low -= FIRST_QTR;                   /* Subtract offset to middle*/
			high -= FIRST_QTR;
		    }
		    else
			break;                             /* Otherwise exit loop.     */
	    low = 2*low;
	    high = 2*high+1;                        /* Scale up code range.     */
	    value = 2*value+input_bit();            /* Move in next input bit.  */
	}
	return symbol;
    }

/* THE ADAPTIVE SOURCE MODEL */

/* INITIALIZE THE MODEL. */

    private void start_model() {
	int i;
	for (i = 0; i<NO_OF_CHARS; i++) {           /* Set up tables that       */
	    char_to_index[i] = i+1;                 /* translate between symbol */
	    index_to_char[i+1] = i; /* indexes and characters.  */
	}
	for (i = 0; i<=NO_OF_SYMBOLS; i++) {        /* Set up initial frequency */
	    freq[i] = 1;                            /* counts to be one for all */
	    cum_freq[i] = NO_OF_SYMBOLS-i;          /* symbols.                 */
	}
	freq[0] = 0;                                /* Freq[0] must not be the  */
    }                                               /* same as freq[1].         */


/* UPDATE THE MODEL TO ACCOUNT FOR A NEW SYMBOL. */

    private void update_model(int symbol) {
	int i;                      /* New index for symbol                     */
	if (cum_freq[0]==MAX_FREQUENCY) {           /* See if frequency counts  */
	    int cum;                                /* are at their maximum.    */
	    cum = 0;
	    for (i = NO_OF_SYMBOLS; i>=0; i--) {    /* If so, halve all the     */
		freq[i] = (freq[i]+1)/2;            /* counts (keeping them     */
		cum_freq[i] = cum;                  /* non-zero).               */
		cum += freq[i];
	    }
	}
	for (i = symbol; freq[i]==freq[i-1]; i--) ; /* Find symbol's new index. */

	if (i<symbol) {
	    int ch_i, ch_symbol;
	    ch_i = index_to_char[i];                /* Update the translation   */
	    ch_symbol = index_to_char[symbol];      /* tables if the symbol has */
	    index_to_char[i] = ch_symbol; /* moved.             */
	    index_to_char[symbol] = ch_i;
	    char_to_index[ch_i] = symbol;
	    char_to_index[ch_symbol] = i;
	}
	freq[i] += 1;                               /* Increment the frequency  */
	while (i>0) {                               /* count for the symbol and */
	    i -= 1;                                 /* update the cumulative    */
	    cum_freq[i] += 1;                       /* frequencies.             */
	}
    }

    public static void main(String[] args) throws IOException {
	InputStream a = new ArithInputStream(System.in);
        int t;
        while((t = a.read()) != -1)
            System.out.write(t);
        System.out.close();
    }
}


