/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Output stream which transforms its data via arithmetic encoding.
 *
 * This implementation is based on ideas of <a
 * href="http://www.dogma.net/markn/articles/bwt/bwt.htm">c source
 * code from Mark Nelson</a>.
 *
 * This stream performs an order-0 adaptive arithmetic encoding
 * function. Based on C code from 1987 CACM article by Witten, Neal,
 * and Cleary.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class ArithOutputStream extends FilterOutputStream implements ArithConstants {

/* TRANSLATION TABLES BETWEEN CHARACTERS AND SYMBOL INDEXES. */

    private int[] char_to_index = new int[NO_OF_CHARS];	/* To index from character          */
    private int[] index_to_char = new int[NO_OF_SYMBOLS+1];	/* To character from index    */

/* ADAPTIVE SOURCE MODEL */

    private int[] freq = new int[NO_OF_SYMBOLS+1];	/* Symbol frequencies                       */
    private int[] cum_freq = new int[NO_OF_SYMBOLS+1];	/* Cumulative symbol frequencies            */

    private boolean writeEOF;
    /**
     * Initialize and set output stream. */
    public ArithOutputStream(OutputStream out) {
	this(out, false);
    }

    /**
     * Initialize and set output stream.
     *
     * @param out stream to write to
     * @param writeEOF write special EOF symbol. Set this to false if
     * you know exactly when to stop reading ... */
    public ArithOutputStream(OutputStream out, boolean writeEOF) {
        super(out);

	this.writeEOF = writeEOF;

	start_model();                              /* Set up other modules.    */
	start_outputing_bits();
	start_encoding();
    }

/* BIT OUTPUT ROUTINES. */

/* THE BIT BUFFER. */

    private int buffer;                     /* Bits buffered for output                 */
    private int bits_to_go;                 /* Number of bits free in buffer            */


/* INITIALIZE FOR BIT OUTPUT. */

    private void start_outputing_bits() {
	buffer = 0;                                 /* Buffer is empty to start */
	bits_to_go = 8;                             /* with.                    */
    }

/* OUTPUT A BIT. */

    private final void output_bit(int bit) throws IOException {
	buffer >>= 1;
	if(bit != 0)
	    buffer |= 0x80;      /* Put bit in top of buffer.*/
	bits_to_go -= 1;
	if(bits_to_go == 0) {	/* Output buffer if it is   */
	    out.write(buffer);	/* now full.                */
	    bits_to_go = 8;
	}
    }

/* FLUSH OUT THE LAST BITS. */

    private void done_outputing_bits() throws IOException {
	out.write(buffer >> bits_to_go);
    }

/* CURRENT STATE OF THE ENCODING. */

    private long low, high;	/* Ends of the current code region          */
    private long bits_to_follow;	/* Number of opposite bits to output after  */
                                /* the next bit.                            */
/* OUTPUT BITS PLUS FOLLOWING OPPOSITE BITS. */

    private void bit_plus_follow(int bit) throws IOException {
	output_bit(bit);                            /* Output the bit.          */
	while (bits_to_follow>0) {
	    output_bit(bit==1?0:1);                       /* Output bits_to_follow    */
	    bits_to_follow -= 1;                    /* opposite bits. Set       */
	}                                           /* bits_to_follow to zero.  */
    }

    /**
     * Encode and write byte.
     *
     * @param b byte to write
     * @exception IOException */
    public void write(int b) throws IOException {
        int symbol = char_to_index[b&0xff];	/* Translate to an index.   */
        encode_symbol(symbol,cum_freq);         /* Encode that symbol.      */
        update_model(symbol);                   /* Update the model.        */
    }

    /**
     * Encode and write byte array. The data is written byte for byte
     * with write(int) to the underlying output stream.
     *
     * @param b byte array to write
     * @param off starting offset
     * @param len number of bytes to write
     * @exception IOException */
    public void write(byte[] b, int off, int len) throws IOException {
	for(int i = off; i < off+len; i++)
	    write(ByteUtils.unsignedByte2Int(b[i]));
    }

    /**
     * Transforms the cached data and writes it to the underlying
     * output stream, which is flushed, too.
     *
     * @exception IOException. */
    public void flush() throws IOException {
	if(writeEOF)
	    encode_symbol(EOF_SYMBOL,cum_freq);         /* Encode the EOF symbol.   */
	done_encoding();                            /* Send the last few bits.  */
	done_outputing_bits();
        out.flush();

	start_model();                              /* Set up other modules.    */
	start_outputing_bits();
	start_encoding();
    }

    /**
     * Transforms the cached data and writes it to the underlying
     * output stream, which is closed, too.
     *
     * @exception IOException. */
    public void close() throws IOException {
	if(writeEOF)
	    encode_symbol(EOF_SYMBOL,cum_freq);         /* Encode the EOF symbol.   */
	done_encoding();                            /* Send the last few bits.  */
	done_outputing_bits();

        out.close();
    }

/* ARITHMETIC ENCODING ALGORITHM. */


/* START ENCODING A STREAM OF SYMBOLS. */

    private void start_encoding() {
	low = 0;                                    /* Full code range.         */
	high = TOP_VALUE;
	bits_to_follow = 0;                         /* No bits to follow next.  */
    }


/* ENCODE A SYMBOL. */

    private void encode_symbol(int symbol, int[] cum_freq) throws IOException {
	long range;                 /* Size of the current code region          */
	range = (long)(high-low)+1;
	high = low +                                /* Narrow the code region   */
	    (range*cum_freq[symbol-1])/cum_freq[0]-1; /* to that allotted to this */
	low = low +                                 /* symbol.                  */
	    (range*cum_freq[symbol])/cum_freq[0];
	for (;;) {                                  /* Loop to output bits.     */
	    if (high<HALF) {
		bit_plus_follow(0);                 /* Output 0 if in low half. */
	    }
	    else
		if (low>=HALF) {                   /* Output 1 if in high half.*/
		    bit_plus_follow(1);
		    low -= HALF;
		    high -= HALF;                       /* Subtract offset to top.  */
		}
		else
		    if (low>=FIRST_QTR                 /* Output an opposite bit   */
			&& high<THIRD_QTR) {              /* later if in middle half. */
			bits_to_follow += 1;
			low -= FIRST_QTR;                   /* Subtract offset to middle*/
			high -= FIRST_QTR;
		    }
		    else
			break;                             /* Otherwise exit loop.     */
	    low = 2*low;
	    high = 2*high+1;                        /* Scale up code range.     */
	}
    }


/* FINISH ENCODING THE STREAM. */

    private void done_encoding() throws IOException {
	bits_to_follow += 1;                        /* Output two bits that     */
	if (low<FIRST_QTR)
	    bit_plus_follow(0);      /* select the quarter that  */
	else
	    bit_plus_follow(1);                    /* the current code range   */
    }                                               /* contains.                */


/* THE ADAPTIVE SOURCE MODEL */

/* INITIALIZE THE MODEL. */

    private void start_model() {
	int i;
	for (i = 0; i<NO_OF_CHARS; i++) {           /* Set up tables that       */
	    char_to_index[i] = i+1;                 /* translate between symbol */
	    index_to_char[i+1] = i; /* indexes and characters.  */
	}
	for (i = 0; i<=NO_OF_SYMBOLS; i++) {        /* Set up initial frequency */
	    freq[i] = 1;                            /* counts to be one for all */
	    cum_freq[i] = NO_OF_SYMBOLS-i;          /* symbols.                 */
	}
	freq[0] = 0;                                /* Freq[0] must not be the  */
    }                                               /* same as freq[1].         */


/* UPDATE THE MODEL TO ACCOUNT FOR A NEW SYMBOL. */

    private void update_model(int symbol) {
	int i;                      /* New index for symbol                     */
	if (cum_freq[0]==MAX_FREQUENCY) {           /* See if frequency counts  */
	    int cum;                                /* are at their maximum.    */
	    cum = 0;
	    for (i = NO_OF_SYMBOLS; i>=0; i--) {    /* If so, halve all the     */
		freq[i] = (freq[i]+1)/2;            /* counts (keeping them     */
		cum_freq[i] = cum;                  /* non-zero).               */
		cum += freq[i];
	    }
	}

	for (i = symbol; freq[i]==freq[i-1]; i--) ; /* Find symbol's new index. */

	if (i<symbol) {
	    int ch_i, ch_symbol;
	    ch_i = index_to_char[i];                /* Update the translation   */
	    ch_symbol = index_to_char[symbol];      /* tables if the symbol has */
	    index_to_char[i] = ch_symbol;           /* moved.                   */
	    index_to_char[symbol] = ch_i;
	    char_to_index[ch_i] = symbol;
	    char_to_index[ch_symbol] = i;
	}
	freq[i] += 1;                               /* Increment the frequency  */
	while (i>0) {                               /* count for the symbol and */
	    i -= 1;                                 /* update the cumulative    */
	    cum_freq[i] += 1;                       /* frequencies.             */
	}
    }

    public static void main(String[] args) throws IOException {
	OutputStream a = new ArithOutputStream(System.out);
        int t;
        while((t = System.in.read()) != -1)
            a.write(t);
        a.close();
    }
}
