/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Filtering output stream which transforms its data with the
 * Burrows-Wheeler Transformation.  For more information see <a
 * href="http://gatekeeper.dec.com/pub/DEC/SRC/research-reports/abstracts/src-rr-124.html">M.
 * Burrows and D. J. Wheeler: A Block-sorting Lossless Data
 * Compression Algorithm</a>.
 *
 * This implementation transforms a whole array if you feed it via the
 * <tt>write(byte[])</tt> method, otherwise it caches the input until
 * <tt>flush()</tt> (or <tt>close()</tt>) is called and then
 * transforms the accummulated data. You can write multiple
 * independent BWT data blocks in order to keep them small, the length
 * of the block precedes the output and the index is written after it,
 * so the output is 8 bytes per block (i.e. <tt>write(byte[])</tt> or
 * <tt>flush()</tt> call) bigger.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class BWTOutputStream extends FilterOutputStream {

    private ByteArrayOutputStream buf = null;

    /**
     * Creates output stream writing to out.
     */
    public BWTOutputStream(OutputStream out) {
	super(out);
    }

    private final static class ArrayIndex {
	int i;

	ArrayIndex(int i) {
	    this.i = i;
	}
    }

    /**
     * Comparator which sorts indices according their "wrapped"
     * strings they point to.  */
    private final static class BoundedCompare implements Comparator {
	final byte[] b;
	final int off;
	final int len;
	
	BoundedCompare(byte[] b, int off, int len) {
	    this.b = b;
	    this.off = off;
	    this.len = len;
	}

	public final int memcmpmod(int s1, int s2) {
	    int i = 0;
	    int t = 0;
	    // Annahme: len > 0
	    while (t == 0 && i++ < len)
		// unsignedByte2Int hier "handgeinlined"
		t = (((int) b[off + (s1++ % len)]) & 0xff) - (((int) b[off + (s2++ % len)]) & 0xff);
		//		t = ByteUtils.unsignedByte2Int(b[off + (s1++ % len)]) - ByteUtils.unsignedByte2Int(b[off + (s2++ % len)]);

	    if (t < 0)
		return -1;
	    else
		if (t > 0)
		    return 1;
                else
		    if (s1 < s2)
			return -1;
		    else
			return 1;
	    // Annahme: s1 != s2
	}

	public final int memcmpmod0(int i1, int i2) {
	    int s1 = i1, s2 = i2;
	    int i = 0, t = 0;
	    // Annahme: len > 0, s1!=s2
	    if(s1 < s2) {
		while (t == 0 && s2 < len)
		    t = (b[s1++]&0xff) - (b[s2++]&0xff);
		if(t == 0) {
		    s2 = 0;
		    while (t == 0 && s1 < len)
			t = (b[s1++]&0xff) - (b[s2++]&0xff);
		    if(t == 0) {
			s1 = 0;
			while (t == 0 && s1 < i1)
			    t = (b[s1++]&0xff) - (b[s2++]&0xff);
		    }
		}
	    } else {
		while (t == 0 && s1 < len)
		    t = (b[s1++]&0xff) - (b[s2++]&0xff);
		if(t == 0) {
		    s1 = 0;
		    while (t == 0 && s2 < len)
			t = (b[s1++]&0xff) - (b[s2++]&0xff);
		    if(t == 0) {
			s2 = 0;
			while (t == 0 && s2 < i2)
			    t = (b[s1++]&0xff) - (b[s2++]&0xff);
		    }
		}
	    }
	    if (t < 0)
		return -1;
	    else
		if (t > 0)
		    return 1;
		else
		    if (s1 < s2)
			return -1;
		    else
			return 1;
	}

	public final int intCompare(int i1, int i2) {
	    if(i1 == i2)
		return 0;
	    return (off==0)?memcmpmod0(i1, i2):memcmpmod(i1, i2);
	}

	public final int compare(Object o1, Object o2) {
	    if(o1.equals(o2))
		return 0;
	    return memcmpmod(((ArrayIndex)o1).i, ((ArrayIndex)o2).i);
        }
    }

    private BoundedCompare comp;

    /**
     * Quicksort, s. Manber S. 134/5, abgewandelt in einer Methode nach
     * Sedgewick S. 148*/
    private final void sort(int[] f, BoundedCompare comp) {
	this.comp = comp;
	quicksort(f, 0, f.length-1);
    }

    private final void quicksort(int[] f, int l, int r) {
	int i, j;
	int v;      // Pivot
	int t;      // temporr
	if(r > l) {
	    v = f[r]; // rechts als Pivot
	    i = l-1;
	    j = r;
	    for(;;) {
		while(comp.intCompare(f[++i], v) < 0) ;
		while(comp.intCompare(f[--j], v) > 0 && j > 0) ;
		if(i >= j)
		    break;
		t = f[i];
		f[i] = f[j];
		f[j] = t;
	    }
	    t = f[i];
	    f[i] = f[r];
	    f[r] = t;
	    quicksort(f, l, i-1);
	    quicksort(f, i+1, r);
	}
    }

    /**
     * Writes a byte. <strong>Attention:</strong> the byte isn't
     * written to the underlying output stream until a call of flush()
     * (or close()). Don't mix it with write(byte[]) without flush(),
     * othwise the output will be mixed up.
     *
     * @param b byte to write
     * @exception IOException */
    public void write(int b) throws IOException {
	if(buf == null)
	    buf = new ByteArrayOutputStream();
	buf.write(b);
    }

    /**
     * Transforms a byte array with BWT and writes it (together with 8
     * bytes of other data) to the underlying output
     * stream. Don't mix it with write(int) without flush(),
     * othwise the output will be mixed up.
     *
     * @param b byte array to write
     * @param off starting offset
     * @param len length of data
     * @exception IOException. */
    public void write(byte[] b, int off, int len) throws IOException {

	out.write(len >> 24 & 0xff);
	out.write(len >> 16 & 0xff);
	out.write(len >> 8 & 0xff);
	out.write(len & 0xff);

        int i;
	/*
	SortedSet p = new TreeSet(new BoundedCompare(b, off, len));
        for (i=0; i < len; i++ )
            p.add(new ArrayIndex(i));

	Iterator ii = p.iterator();
        int first = 0;
	int t;
        for (i = 0; ii.hasNext() ; i++) {
	    t = (((ArrayIndex)(ii.next())).i);
	    if(t == 0)
		first = i;
	    out.write(ByteUtils.unsignedByte2Int(b[off+((len+t-1)%len)]));
	}
	*/

	/*
	ArrayIndex[] ai = new ArrayIndex[len];

	for(i = 0; i < len; i++)
	    ai[i] = new ArrayIndex(i);

	Arrays.sort(ai, new BoundedCompare(b, off, len));

        int first = 0;
	int t;

	for(i = 0; i < len; i++) {
	    t = ai[i].i;
	    if(t == 0)
		first = i;
	    out.write(ByteUtils.unsignedByte2Int(b[off+((len+t-1)%len)]));
	}
	*/

	/**/
	int[] ai = new int[len];

	for(i = 0; i < len; i++)
	    ai[i] = i;

	sort(ai, new BoundedCompare(b, off, len));

        int first = 0;
	int t;

	for(i = 0; i < len; i++) {
	    t = ai[i];
	    if(t == 0)
		first = i;
	    out.write(ByteUtils.unsignedByte2Int(b[off+((len+t-1)%len)]));
	}
	/**/

	//	System.err.println(first);

	out.write(first >> 24 & 0xff);
	out.write(first >> 16 & 0xff);
	out.write(first >> 8 & 0xff);
	out.write(first & 0xff);

	//	System.err.println(off + " " + len +  " fertig");

    }

    /**
     * Transforms the cached data (via write(int)) and writes it to
     * the underlying output stream, which is flushed, too.
     *
     * @exception IOException. */
    public void flush() throws IOException {
	if(buf != null) {
	    buf.writeTo(this);
	    buf.reset();
	}
	out.flush();
    }

    /**
     * Transforms the cached data (via write(int)) and writes it to
     * the underlying output stream, which is closed, too.
     *
     * @exception IOException. */
    public void close() throws IOException {
	if(buf != null) {
	    buf.writeTo(this);
	    buf.reset();
	}
	out.close();
    }

    public static void main(String[] args) throws IOException {
	BWTOutputStream bwt = new BWTOutputStream(System.out);

	/*	bwt.write("1abacb\n".getBytes(), 1, 6);
	bwt.write("barrakuda\n".getBytes());

	bwt.write('a');
	bwt.write('b');
	bwt.write('a');
	bwt.write('c');
	bwt.write('b');
	bwt.write('\n');
	bwt.close();*/

	int t;
	while((t = System.in.read()) != -1)
	    bwt.write(t);
	//	bwt.flush();
	bwt.close();
    }
}
