/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Input stream which transforms its data via
 * Run-Length-Decoding.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class RLEInputStream extends FilterInputStream {

    private int count = 0;
    private int lastchar;

    private static final int COUNT_MAX = 254;
    private static final int ESCAPE_CHAR = 0xff;

    /**
     * Initialize and set input stream. */
    public RLEInputStream(InputStream in) {
	super(in);
    }

    /**
     * Reads and transforms a byte from the underlying input
     * stream.
     *
     * @return tranformed byte or -1 for end of file
     * @exception IOException
     */
    public int read() throws IOException {
	int t;

	if(count == 0)
	    t = in.read();
	else {
	    count--;
	    return lastchar;
	}

	if(t == -1)
	    return t;

	if(t == ESCAPE_CHAR) {
	    int newt = in.read();
	    if(newt == -1)
		throw new IOException("RLE: File format error");
	    if(newt == ESCAPE_CHAR)
		return lastchar = ESCAPE_CHAR;
	    else {
		count = newt;
		if(count == COUNT_MAX)
		    count--;
		lastchar = in.read();
		if(lastchar == -1)
		    throw new IOException("RLE: File format error");
	    }
	    return lastchar;
	} else
	    return lastchar = t;
    }

    /**
     * Reads and transforms a byte array from the underlying input
     * stream. The data is read byte for byte via read() from the
     * underlying input stream.
     *
     * @param b where to write the bytes to
     * @param off starting offset
     * @param len how many bytes to read
     * @return number of bytes bytes actually read
     * @exception IOException */
    public int read(byte[] b, int off, int len) throws IOException {
	int t, i;
        for(i = off; i < off+len; i++)
            if((t = read()) == -1)
		break;
	    else
		b[i] = (byte)t;
	return i;
    }

    /**
     * Do nothing.
     */
    public void mark(int readlimit) {
    }

    /**
     * Do nothing.
     */
    public void reset() throws IOException {
    }

    /**
     * @return false
     */
    public boolean markSupported() {
	return false;
    }

    /**
     * Skips bytes.
     *
     * @param n how many bytes to skip.
     * @return number of bytes actually skipped
     * @exception IOException */
    public long skip(long n) throws IOException {
	long i;
	int t;
	for(i = 0; i < n; i++)
	    if((t = read()) == -1)
		break;
	return i;
    }

    public static void main(String[] args) throws IOException {
	InputStream a = new RLEInputStream(System.in);
        int t;
        while((t = a.read()) != -1)
            System.out.write(t);
        System.out.close();
    }
}
