/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Output stream which transforms its data via
 * Run-Length-Encoding.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class RLEOutputStream extends FilterOutputStream {

    private int count = 0;
    private int lastchar;

    private static final int COUNT_MAX = 254;
    private static final int ESCAPE_CHAR = 0xff;

    boolean first = true;

    /**
     * Initialize and set output stream. */
    public RLEOutputStream(OutputStream out) {
        super(out);
    }

    private void writeByte(int b) throws IOException {
	if(b == ESCAPE_CHAR)
	    out.write(ESCAPE_CHAR);
	out.write(b);
    }

    private void writeRLE() throws IOException {
	if((count > 1) || (lastchar == ESCAPE_CHAR)) {
	    out.write(ESCAPE_CHAR);
	    out.write(count);
	    out.write(lastchar);
	} else
	    for(int i = 0; i <= count; i++)
		writeByte(lastchar);
	count = 0;
    }
    
    /**
     * Encode and write byte.
     *
     * @param b byte to write
     * @exception IOException */
    public void write(int b) throws IOException {
	if(first) {
	    lastchar = b;
	    first = false;
	    return;
	}
	if(b == lastchar) {
	    count++;
	    if(count == COUNT_MAX)
		writeRLE();
	} else
	    if(count <= 0)
		writeByte(lastchar);
	    else
		writeRLE();
	lastchar = b;	
    }

    /**
     * Encode and write byte array. The data is written byte for byte
     * with write(int) to the underlying output stream.
     *
     * @param b byte array to write
     * @param off starting offset
     * @param len number of bytes to write
     * @exception IOException */
    public void write(byte[] b, int off, int len) throws IOException {
	for(int i = off; i < off+len; i++)
	    write(ByteUtils.unsignedByte2Int(b[i]));
    }

    /**
     * Transforms the cached data and writes it to the underlying
     * output stream, which is flushed, too.
     *
     * @exception IOException. */
    public void flush() throws IOException {
	if(!first)
	    if(count > 0)
		writeRLE();
	    else
		writeByte(lastchar);

	out.flush();
    }

    /**
     * Closes the underlying output stream.
     *
     * @exception IOException. */
    public void close() throws IOException {
	if(!first)
	    if(count > 0)
		writeRLE();
	    else
		writeByte(lastchar);

	out.close();
    }

    public static void main(String[] args) throws IOException {
	OutputStream a = new RLEOutputStream(System.out);
        int t;
        while((t = System.in.read()) != -1)
            a.write(t);
        a.close();
    }
}
