/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.math;

import java.util.*;
import java.io.*;
import java.text.*;
import picture.math.SymMat;
import picture.PictureException;

/**
 * Karhunen-Loeve-Tranformation */
public class KLT {
    private double[][] m;
    private int size;
    private EigenRaum[] ers;

    private class EigenRaum implements Comparable {
	double ew;
	int evindex;

	EigenRaum(double ew, int evindex) {
	    this.ew = ew;
	    this.evindex = evindex;
	}
	
	public int compareTo(Object o) throws ClassCastException {
	    if(!(o instanceof EigenRaum))
		throw new ClassCastException();

	    EigenRaum er = (EigenRaum)o;

	    if(Math.abs(ew) == Math.abs(er.ew))
		return 0;
	    else
		return (Math.abs(ew) > Math.abs(er.ew)) ? -1 : 1;
	}
    }
    

    /**
     * Returns the Autocovariance matrice of the double arrays with
     * which it was called. */
    private static double[][] makeAutoCovarianceMatrice(double[][] vec) {
	if(vec == null || vec.length == 0)
	    return null;

	double[][] out = new double[vec[0].length+1][vec[0].length+1];

	final int n = vec.length;

	for(int i = 1; i <= vec[0].length; i++)
	    for(int j = 1; j <= vec[0].length; j++)
		out[i][j] = 0.0;

	for(int k = 0; k < n; k++) {
	    double[] x = vec[k];
	    for(int i = 0; i < vec[0].length; i++)
		for(int j = 0; j < vec[0].length; j++)
		    out[i+1][j+1] += x[i]*x[j];
	}

	for(int i = 1; i <= vec[0].length; i++)
	    for(int j = 1; j <= vec[0].length; j++)
		out[i][j] /= n;

	return out;

    }

    /**
     * Inititializes a Karhunen-Loeve-Object. Therefore it needs an
     * array of double[]s, which are used to build the basis
     * transformation matrice. It assumes that the double[]s are
     * normalized, i.e. their mean value is {0.0, 0.0, ..., 0.0}.
     *
     * @param vecs array of double[]s which holds the normalized
     * vectors of the input. */
    public KLT(double[][] vecs) throws PictureException {
	size = vecs[0].length;
	m = makeAutoCovarianceMatrice(vecs);

	//	SymMat.printMatrice(m);
	//	System.err.println();
	
	double[] d = new double[size+1];
	double[] e = new double[size+1];

	SymMat.tred2(m, size, d, e);
	SymMat.tqli(d, e, size, m);

	//	SymMat.printMatrice(m);
	//	System.err.println();
	
	ers = new EigenRaum[size];

	for(int i = 0; i < size; i++)
	    ers[i] = new EigenRaum(d[i+1], i+1);

	Arrays.sort(ers);

	//       	for(int i = 0; i < size; i++)
	//	    System.err.println(d[i+1] + " " + ers[i].ew + " " + ers[i].evindex);
	//	System.err.println();
    }

    /**
     * Returns the Eigenvalues of this transformation by descending
     * magnitude. */
    public double[] getEigenValues() {
        double[] ev = new double[ers.length];

        for (int i = 0; i < ev.length; i++)
            ev[i] = ers[i].ew;

        return ev;	
    }
    
    /**
     * Returns the Eigenvectors of this transformation.
     */
    public double[][] getEigenVectors() {
        double[][] ev = new double[ers.length][size];

        for (int i = 0; i < ev.length; i++)
	    for (int j = 0; j < ev[0].length; j++)
		ev[i][j] = m[ers[i].evindex][j+1];

        return ev;	
    }
    
    /**
     * Tranforms a vector. */
    public double[] transform(double[] x) {
	if(x.length != size)
	    return x;

	double[] erg = new double[size];

	for(int i = 0; i < size; i++)
	    erg[i] = 0;

	for(int i = 0; i < size; i++)
	    for(int j = 0; j < size; j++)
		erg[i] += m[j+1][ers[i].evindex]*x[j];

	return erg;
    }

    /**
     * Inverse transform a vector. */
    public double[] inverseTransform(double[] x) {
	if(x.length != size)
	    return x;

	double[] erg = new double[size];

	for(int i = 0; i < size; i++)
	    erg[i] = 0;

	for(int i = 0; i < size; i++)
	    for(int j = 0; j < size; j++)
		erg[i] += m[i+1][ers[j].evindex]*x[j];

	return erg;
    }

    /**
     * Inverse transform a vector after adjusting weights. */
    public double[] inverseTransform(double[] x, double[] weights) {
	if(x.length != size)
	    return x;

	double[] erg = new double[size];

	for(int i = 0; i < size; i++)
	    erg[i] = 0;

	for(int i = 0; i < size; i++)
	    for(int j = 0; j < size; j++)
		erg[i] += m[i+1][ers[j].evindex]*x[j]*weights[j];

	return erg;
    }

    /**
     * Outputs the transformation base to stderr.
     */
    public void printBase() {
	for(int x = 0; x < size; x++) {
	    for(int y = 0; y < size; y++)
		System.err.print(format(m[x+1][ers[y].evindex]) + " ");
	    System.err.println();
	}
	System.err.println();
    }
    
    /**
     * Normalizes a bunch of vectors, it returns a normalized version
     * and the mean values in the array mean */
    public static double[][] normalize(double[][] vecs, double[] mean) {
	if(vecs == null || vecs.length == 0 || mean.length < vecs[0].length)
	    return null;

	double[][] out = new double[vecs.length][vecs[0].length];

	int n = vecs.length;

	for(int i = 0; i < vecs[0].length; i++)
	    mean[i] = 0;

	for(int i = 0; i < vecs[0].length; i++)
	    for(int k = 0; k < n; k++)
		mean[i] += vecs[k][i];

	for(int i = 0; i < vecs[0].length; i++)
	    mean[i] /= n;

	for(int i = 0; i < vecs[0].length; i++)
	    for(int k = 0; k < n; k++)
		out[k][i] = vecs[k][i] - mean[i];

	return out;
    }

    /**
     * Return the normalization of the given vector v with the mean
     * vector. */
    public static double[] normal(double[] v, double[] mean) {
	double[] out = new double[v.length];

	for(int i = 0; i < v.length; i++)
	    out[i] = v[i] - mean[i];

	return out;
    }

    /**
     * Return the denormalization of the given vector v with the mean
     * vector. */
    public static double[] denormal(double[] v, double[] mean) {
	double[] out = new double[v.length];

	for(int i = 0; i < v.length; i++)
	    out[i] = v[i] + mean[i];

	return out;
    }

    private static String format(double x, int vor, int nach) {
        char[] sp = new char[vor];
        for (int i = 0; i < vor; i++)
            sp[i] = ' ';
        String spaces = new String(sp);
        DecimalFormat f = new DecimalFormat();
        f.setMinimumFractionDigits(nach);
        StringBuffer buf = new StringBuffer();
        FieldPosition fpos = new FieldPosition(NumberFormat.INTEGER_FIELD);
        f.format(x, buf, fpos);
        int anzSp = vor - (fpos.getEndIndex() - fpos.getBeginIndex());
        if (x < 0)
            --anzSp;
        if (anzSp < 0)
            return buf.toString();
        else
            return spaces.substring(0, anzSp) + buf.toString();
    }

    private static final String format(double x) {
	return format(x, 10, 5);
    }

    public static void main(String[] args) {
	try {
	    //	    KLT klt = new KLT(new double[][] { { 1.0, 2.0 } , { 1.0, 0.0 } , { 1.0, 0.0 }, { 1.0, 0.0 } });
	    //	    KLT klt = new KLT(new double[][] { { 1.0, 2.0 } });
	    //	    double[] a = { 1.0, 2.0 };

	    double[] a = new double[50];
	    // Raum fr 10 50-dimensionale Vektoren
	    double[][] t = new double[10][50];

	    Random r = new Random();

	    // Initialisieren der Vektoren
	    for(int k = 0; k < t.length; k++)
		for(int i = 0; i < t[0].length; i++)
		    t[k][i] = r.nextGaussian();

	    KLT klt = new KLT(t);

	    // KLT ...
	    a = klt.transform(t[0]);

	    for(int i = 0; i < a.length; i++)
		System.out.println(format(i, 3, 0) + ": " + format(a[i]));
	    System.out.println();

	    // Wegschmeien der zu den kleinen EW gehrigen Werte
	    for(int i = 10; i < a.length; i++)
		a[i] = 0.0;

	    // Rcktransformieren
	    a = klt.inverseTransform(a);

	    for(int i = 0; i < a.length; i++)
		System.out.println(format(i, 3, 0) + ": " + format(t[0][i]) + " " + format(a[i]) + " " + format(t[0][i]-a[i]));
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }
    
}
